"""
This module will contain data processing routines that operate on measured or simulated waveforms.
"""

from .wave import align_waves, get_effective_response
from .trace_handler import TraceHandler, MultiTraceHandler
__all__ = [
    "align_waves",
    "get_effective_response",
    "TraceHandler",
    "MultiTraceHandler"
]



"""
Tools from the Attosecond science group at the Max Planck Institute of Quantum Optics, a.k.a. [Attoworld](https://www.attoworld.de)
"""
from . import file
from . import numeric
from . import personal
from . import plot
from . import spectrum
from . import wave
from . import attoworld_rs



"""
This module will contain functions for loading the various data formats used across the labs.
"""

from .file_io import load_waves_from_matfile
from .FROG_result import FrogResult
from .interface_simulations import LunaResult
from .profileAndIntensity import profile_analysis

__all__ = ["load_waves_from_matfile","FrogResult","LunaResult","profile_analysis"]



"""
This module will contain numerical tools.
"""

from .numeric import uniform_derivative, interpolate
from .atomic_units import AtomicUnits
from ..attoworld_rs import fornberg_stencil, fwhm, find_first_intercept, find_last_intercept
__all__ = [
    "fornberg_stencil",
    "fwhm",
    "find_first_intercept",
    "find_last_intercept",
    "uniform_derivative",
    "interpolate",
    "AtomicUnits"
]



import numpy as np
import scipy.signal
import h5py

e = 1.602176462e-19
hbar = 1.05457159682e-34
m = 9.1093818872e-31
eps0 = 8.854187817e-12
kB = 1.380650324e-23
c = 299792458

figureSize = [30,12]
tickSize = 48
fontSize = 55
lineWidth = 6
legendFontSize = 41
axisLineWidth = 3

def check_equal_length(*arg):

    n = len(arg[0])
    for v in arg:
        if len(v) != n:
            print(v)
            print('Error: vector size mismatch')
            raise Exception('Vector size mismatch')

def fourier_transform(TimeV, FieldV):   # complex!!!

    freq = np.fft.fftfreq(TimeV.size, d=TimeV[1]-TimeV[0])
    fft = np.fft.fft(FieldV)
    return freq, fft

def inverse_fourier_transform(freq, fullSpectrum):  # complex!!!

    timeV = np.fft.fftfreq(len(freq), freq[1]-freq[0])
    if len(timeV)%2 == 0:
        timeV = np.concatenate((timeV[int(len(timeV)/2):], timeV[:int(len(timeV)/2)]))
    else:
        timeV = np.concatenate((timeV[int((len(timeV)+1)/2):], timeV[:int((len(timeV)+1)/2)]))
    fieldV = np.fft.ifft(fullSpectrum)

    return timeV, fieldV

class LunaResult:
    """Loads and handles the Luna simulation result.

    The result must be in the HDF5 format using the saving option in the Luna.Interface.prop_capillary() function [filepath="..."].
    As opposed to most of the analysis routines here, units are in SI!"""
    def __init__(self, filename):
        self.filename = filename
        self.fieldFT = None
        self.omega = None
        self.stats = None
        self.z = None
        self.grid = None
        self.open_Luna_result(filename)

    def open_Luna_result(self, filename):
        """Opens the Luna result file and loads the data"""
        with h5py.File(filename, 'r') as f:
            data = f
            self.fieldFT = np.array(data['Eω'])
            self.grid = data['grid']
            self.omega = np.array(self.grid['ω'])
            self.z = np.array(data['z'])
            self.stats = data['stats']

    def average_modes(self):
        """Averages the propagation modes in the Luna result file"""
        if len(self.fieldFT.shape) == 3:
            self.fieldFT = np.mean(self.fieldFT, axis=1)

    def select_mode(self, mode: int):
        if len(self.fieldFT.shape) < 3:
            print("WARNING: No mode to select")
        elif mode > self.fieldFT.shape[1] or mode < 0:
            print("WARNING: mode ", mode, " is out of range")
        else:
            self.fieldFT = self.fieldFT[:, mode, :]

    def get_time_field(self, position=None):
        """Get the COMPLEX electric field in time from the Luna result file

        Args:
            position (float): position along the fiber in m. If None, the end of the fiber is used.

        Returns:
            timeV (numpy.ndarray): time axis in seconds
            fieldV (numpy.ndarray): electric field in V/m ( COMPLEX!! -> remember to take the real part to get the field or the absolute value to get the envelope
                if you take the imaginary part, I bear no responsibility for that)
        """
        self.average_modes()
        if position is None:
            position = self.z[-1]
        index = np.argmin(np.abs(self.z - position))
        if position > np.max(self.z) or position < np.min(self.z):
            print("WARNING: position ", position, "m is out of range")
        check_equal_length(self.fieldFT[index], self.omega)
        fieldFFT = np.concatenate((self.fieldFT[index, :], np.conjugate(self.fieldFT[index, :][::-1])*0))
        freq = np.concatenate((self.omega, -self.omega[::-1])) / 2 / np.pi
        timeV, fieldV = inverse_fourier_transform(freq, fieldFFT)
        return timeV, fieldV

    def get_wavelength_spectrum(self, position=None):
        """Get the spectrum from the Luna result file (|FFT|^2 * (2 * pi * c / λ^2))

        Args:
            position (float): position along the fiber in m. If None, the end of the fiber is used.

        Returns:
            wvl (numpy.ndarray): wavelength axis in m
            wvlSpectrum (numpy.ndarray): electric field spectrum in V/m ( COMPLEX !! )
        """
        self.average_modes()
        if position is None:
            position = self.z[-1]
        index = np.argmin(np.abs(self.z - position))
        if position > np.max(self.z) or position < np.min(self.z):
            print("WARNING: position ", position, "m is out of range")
        wvl = 2 * np.pi * c / self.omega[::-1]
        wvlSpectrum = np.abs(self.fieldFT[index, ::-1])**2 * (2 * np.pi * c/wvl**2 )
        return wvl, wvlSpectrum


    def get_spectral_phase(self, position=None):
        """Get the spectral phase from the Luna result file

        Args:
            position (float): position along the fiber in m. If None, the end of the fiber is used.

        Returns:
            wvl (numpy.ndarray): wavelength axis in m
            phase (numpy.ndarray): spectral phase in rad
        """
        self.average_modes()
        if position is None:
            position = self.z[-1]
        index = np.argmin(np.abs(self.z - position))
        if position > np.max(self.z) or position < np.min(self.z):
            print("WARNING: position ", position, "m is out of range")
        wvl = 2 * np.pi * c / self.omega[::-1]
        phase = np.angle(self.fieldFT[index, ::-1])
        return wvl, phase






import numpy as np
from matplotlib import pyplot as plt
import pandas
import scipy.optimize as opt
import scipy.signal

e = 1.602176462e-19
hbar = 1.05457159682e-34
m = 9.1093818872e-31
eps0 = 8.854187817e-12
kB = 1.380650324e-23
c = 299792458


def get_fwhm(t, x, no_envelope: bool = False):
    """
    This function calculates the full-width-at-half-maximum:
    FWHM is the full-width of the intensity profile I(t) = |A(t)|^2 at half-maximum, where A(t) is the envelope of the signal.

    Args:
    x: signal in time domain vector
    t: time vector

    RETURNS:
    fwhm: fwhm_val[0]

    if no_envelope == True it computes the FWHM without taking the square modulus and without computing envelope (assuming input is intensity envelope already)
    """
    if no_envelope and np.min(x) < 0:
        raise ValueError('tried to compute FWHM of an envelope with negative values')
    envelope = np.abs(scipy.signal.hilbert(x) ** 2)
    half_max = np.max(envelope) / 2
    finet = np.linspace(t[0], t[-1], len(t)*50)
    fineEnvelope = np.interp(finet, t, envelope)
    if no_envelope:
        fineEnvelope = np.interp(finet, t, x)
        half_max = np.max(x) / 2

    signs = np.sign(fineEnvelope - half_max)
    fwhm_val = np.abs(finet[np.argwhere(signs == 1)[0]] - finet[np.argwhere(signs == 1)[-1]])
    return fwhm_val[0]

def gaussian(height, center_x, center_y, width_x, width_y, offset):
    """Returns a gaussian function with the given parameters"""
    width_x = float(width_x)
    width_y = float(width_y)
    return lambda x,y: offset + height*np.exp(
                -(((center_x-x)/width_x)**2+((center_y-y)/width_y)**2)/2)

def eliminate_broken_pixels(data):
    """Eliminates the broken pixels of the camera rayCi I was using"""
    data[395:409, 950:1060] = 0
    return data

def moments(data):
    """Returns (height, x, y, width_x, width_y)
    the gaussian parameters of a 2D distribution by calculating its
    moments """
    total = data.sum()
    X, Y = np.indices(data.shape)
    x = (X*data).sum()/total
    y = (Y*data).sum()/total
    col = data[:, int(y)]
    width_x = np.sqrt(np.abs((np.arange(col.size)-x)**2*col).sum()/col.sum())
    row = data[int(x), :]
    width_y = np.sqrt(np.abs((np.arange(row.size)-y)**2*row).sum()/row.sum())
    height = data.max()
    return height, x, y, width_x, width_y, np.min(data)

def moments_peak(data):
    """Returns (height, x, y, width_x, width_y, np.min(data)), that is,
    the gaussian parameters of a 2D distribution by calculating its
    peak and moments"""
    total = data.sum()
    X, Y = np.indices(data.shape)
    maxindex = np.unravel_index(np.argmax(data), data.shape)
    x, y = maxindex[0], maxindex[1]
    col = data[:, int(y)]
    width_x = np.sqrt(np.abs((np.arange(col.size)-x)**2*col).sum()/col.sum())
    row = data[int(x), :]
    width_y = np.sqrt(np.abs((np.arange(row.size)-y)**2*row).sum()/row.sum())
    height = data.max()
    return height, x, y, width_x, width_y, np.min(data)

def fitgaussian(data):
    """Returns (height, x, y, width_x, width_y)
    the gaussian parameters of a 2D distribution found by a fit"""
    params = moments_peak(data)
    errorfunction = lambda p: np.ravel(gaussian(*p)(*np.indices(data.shape)) -
                                 data)
    p, success = opt.leastsq(errorfunction, params)
    return p

def plot_crosssect(data, fitfunct=None):
    maxindex = np.unravel_index(np.argmax(data), data.shape)
    y, x = maxindex[0], maxindex[1]
    Y, X = np.indices(data.shape)

    fig, ax = plt.subplots(1, 2)
    ax[0].plot(X[y], data[y])
    if fitfunct is not None:
        ax[0].plot(X[y], fitfunct(X[y]*0. + y, X[y]))
    ax[1].plot(Y[:,x], data[:,x])
    if fitfunct is not None:
        ax[1].plot(Y[:,x], fitfunct(Y[:,x], Y[:,x]*0. + x))
    fig.show()

def cut_tail(data, radius, pixelsize=1., center=None):
    radius=radius/pixelsize
    if center == None:
        center = np.unravel_index(np.argmax(data), data.shape)
    y, x = center[0], center[1]
    Y, X = np.indices(data.shape)
    cut_data = data.copy()
    cut_data[np.sqrt((Y-y)**2+(X-x)**2) > radius] = 0.
    return cut_data

def cut_trace(t, x, radius, center=None):
    if center == None:
        center = np.argmax(abs(x))
    cut_trace = x.copy()
    cut_trace[abs(t-t[center])>radius] = 0
    return cut_trace

def integral2d(data, pixelsize=1.):
    return data.sum()*pixelsize**2

def trace_integral(t, x):
    dt = np.gradient(t)
    f = x #np.array([(i+j)/2 for (i,j) in zip(x[:-1], x[1:])])
    return (dt*f).sum()

def twoD_Gaussian(xy, amplitude, xo, yo, sigma_x, sigma_y, theta, offset):
    x, y = xy
    xo = float(xo)
    yo = float(yo)
    a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
    b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
    c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
    g = offset + amplitude*np.exp( - (a*((x-xo)**2) + 2*b*(x-xo)*(y-yo)
                            + c*((y-yo)**2)))
    return g.ravel()


def profile_analysis(profile_file, trace_file = None, magnification=1., pixelsize=5., power=20., reprate=4., ROI_diam:int=100, forced_background=None, trace_cutoff_radius = 50., cutoff_gaus_fit_profile = 25.):
    """if trace_file is None, this function only loads, fits and plots the beam profile; if trace_file is given, the function additionally calculates the peak intensity and field.

    ARGUMENTS:
        profile_file: name of the file containing the beam profile; the file (txt or csv) contains the 2D array of the beam profile

    OPTIONAL ARGUMENTS:
        trace_file: None; name of the file containing the trace; the file (txt or csv) contains the time and field arrays as tab-separated columns
        magnification: 1.; ratio of lens-camera/lens-z0 lengths; This is used to give the correct size of the beam profile at z0 in um
        pixelsize: 5. um; pixel size of the camera
        power: 20. mW; power of the laser beam; this is used to calculate the peak intensity
        reprate: 4 kHz; repetition rate of the laser; this is used to calculate the peak intensity
        ROI_diam: 100; diameter of the region of interest in pixels; the center of the ROI is the maximum of the camera data (which is assumed to be not saturated)
        cutoff_gaus_fit_profile: 25. pixels; the 2d gaussian is fitted to the beam profile up to this distance from the peak
        forced_background: None; if not None, the subtracted background is set to this value; otherwise it's computed from the outer ring of the ROI (camera data)
        trace_cutoff_radius: 50. fs; for intensity calculation, the integral of the trace is computed only in the range t[peak]-trace_cutoff_radius < t < t[peak]+trace_cutoff_radius"""

    pixelsize=pixelsize/magnification
    data=pandas.read_csv(profile_file)

    dataval=np.array(data.values)
    #print('eliminating fixed broken pixels rayCi camera')
    #dataval=eliminate_broken_pixels(dataval)
    maxindex = np.unravel_index(np.argmax(dataval), dataval.shape)
    print('diameter of Region Of Interest (ROI): ', ROI_diam, ' pixels')
    dataval=np.array(data.values[maxindex[0]-int(ROI_diam/2):maxindex[0]+int(ROI_diam/2),maxindex[1]-int(ROI_diam/2):maxindex[1]+int(ROI_diam/2)])
    maxindex = np.unravel_index(np.argmax(dataval), dataval.shape)

    background = 0.
    count = 0
    print('calculating background intensity outside a circle whose diameter is 3/4 of the ROI diameter')
    for i, j in np.ndindex(dataval.shape):
        if (i-maxindex[0])**2 + (j-maxindex[1])**2 > (ROI_diam*3/8)**2:
            background += dataval[i,j]
            count += 1
    background /= count

    if forced_background is not None:
        background = forced_background
        print('subtracting background intensity: ', background)
    else:
        print('subtracting background intensity: ', background)

    dataval=np.maximum(dataval-background, dataval*0.)
    maxval = np.max(dataval)

    # gaussian fit
    param = moments_peak(dataval)
    param = fitgaussian(cut_tail(dataval, cutoff_gaus_fit_profile))
    X, Y = np.indices(dataval.shape)
    fitted_funct = gaussian(*param)(X, Y)
    print('beam w [um], resp. in x and y axis:', 2*param[3]*pixelsize, 2*param[4]*pixelsize)
    print('beam 1/e^2 diameter [um]: ', 4*param[3]*pixelsize, 4*param[4]*pixelsize)

    if  trace_file is not None:

        trace = pandas.read_csv(trace_file, sep='\t')

        # time in fs, field in a.u.
        intensity = power *1e-3/reprate *1e-3\
                / integral2d(fitted_funct, pixelsize=pixelsize) * np.max(fitted_funct) *1e8 \
                / trace_integral(trace['delay (fs)'], cut_trace(trace['delay (fs)'], trace['field (a.u.)']**2, trace_cutoff_radius)) * np.max(trace['field (a.u.)']**2) *1e15

        intensityFromRawCameraData = power *1e-3/reprate *1e-3\
                / integral2d(dataval[maxindex[0]-int(ROI_diam/2):maxindex[0]+int(ROI_diam/2),maxindex[1]-int(ROI_diam/2):maxindex[1]+int(ROI_diam/2)], pixelsize=pixelsize) * np.max(dataval) *1e8 \
                / trace_integral(trace['delay (fs)'], cut_trace(trace['delay (fs)'], trace['field (a.u.)']**2, trace_cutoff_radius)) * np.max(trace['field (a.u.)']**2) *1e15

        temporal_fwhm = get_fwhm(np.array(trace['delay (fs)']), np.array(trace['field (a.u.)']))
        intensityFromTemporalFWHM = power *1e-3/reprate *1e-3\
                / integral2d(dataval[maxindex[0]-int(ROI_diam/2):maxindex[0]+int(ROI_diam/2),maxindex[1]-int(ROI_diam/2):maxindex[1]+int(ROI_diam/2)], pixelsize=pixelsize) * np.max(dataval) *1e8 \
                / (np.sqrt(2*np.pi)*temporal_fwhm/np.sqrt(8*np.log(2))) *1e15

        print('\nusing the gaussian fit of the beam profile and the integral of the trace^2:\n '
            'peak intensity (sub-cycle, W/cm2): {i:.3e}'.format(**{'i':intensity}))
        peak_field = np.sqrt(intensity*1e4 /c/eps0)
        print('peak field (sub-cycle, V/m): {f:.3e}'.format(**{'f':peak_field}))

        print('\nusing directly camera data, and the integral of the trace^2:\n '
              'peak intensity (sub-cycle, W/cm2): {i:.3e}'.format(**{'i':intensityFromRawCameraData}))
        peak_field = np.sqrt(intensityFromRawCameraData*1e4 /c/eps0)
        print('peak field (sub-cycle, V/m): {f:.3e}'.format(**{'f':peak_field}))

        print('\nusing directly camera data, and the temporal FWHM retrieved from the trace:\n '
              'peak intensity (sub-cycle, W/cm2): {i:.3e}'.format(**{'i':intensityFromTemporalFWHM}))
        peak_field = np.sqrt(intensityFromTemporalFWHM*1e4 /c/eps0)
        print('peak field (sub-cycle, V/m): {f:.3e}'.format(**{'f':peak_field}))

        fig, ax = plt.subplots(1, 1)
        time = trace['delay (fs)']
        field = trace['field (a.u.)']
        i_peak = np.argmax(field**2)
        ax.plot(time[np.abs(time-time[i_peak])<trace_cutoff_radius],
                field[np.abs(time-time[i_peak])<trace_cutoff_radius]**2)
        plt.show()

    fig, ax = plt.subplots(1, 1)
    ax.imshow(dataval)
    plt.show()

    plot_crosssect(dataval, fitfunct=gaussian(*param))


import marimo

__generated_with = "0.13.15"
app = marimo.App(width="medium")


@app.cell
def _():
    import marimo as mo
    import numpy as np
    import attoworld as aw
    import matplotlib.pyplot as plt
    return aw, mo, np, plt


@app.cell
def _(mo):
    mo.md(r"""## Check convergence of numerical derivatives""")
    return


@app.cell
def _(aw, np, plt):
    def convergence_check(order, N_pts):
        x = np.linspace(0.0, 2*np.pi, N_pts+1)[0:-1]
        dx = x[1]-x[0]
        y = np.cos(x)
        y_derivative_analytic = -np.sin(x)
        y_derivative = aw.numeric.uniform_derivative(y,1,order,boundary='periodic')/dx
        return np.max(np.abs(y_derivative - y_derivative_analytic))

    N_pts_range = [16, 32, 64, 128, 256, 512]
    order_range = range(1,6)

    convergence_data = np.array([[convergence_check(_order, _n) for _n in N_pts_range] for _order in order_range])

    lines = plt.loglog(N_pts_range,convergence_data.T)
    for order, line in zip(order_range, lines):
        line.set_label(f"{order}")
    plt.xlabel("Number of grid points")
    plt.ylabel("Mean error")
    plt.ylim(1e-15, 0.1)
    plt.legend()
    aw.plot.showmo()
    return


@app.cell
def _(mo):
    mo.md(r"""## Check interpolate()""")
    return


@app.cell
def _(aw, np, plt):
    def plot_interpolate_test():
        x = np.real(np.linspace(0.0,17.0,24))
        x_fine = np.real(np.linspace(0.0,17.0,1024))
        y = np.sin(x**2/10)
        y_fine = np.sin(x_fine**2/10)
        x2 = np.linspace(0.0,19.0,60)
        plt.plot(x,y,'o')
        plt.plot(x_fine,y_fine)
        plt.plot(x2,aw.numeric.interpolate(x2, x,y, 3, extrapolate=False),'x')

    plot_interpolate_test()
    aw.plot.showmo()
    return


@app.cell
def _(aw, np, plt):
    ix = np.linspace(-5, 5, 15)
    _dx = ix[1]-ix[0]
    iy = np.exp(-ix**2/2)
    plt.plot(iy)

    interpolated_position, interpolated_max = aw.attoworld_rs.find_maximum_location(iy, 3)
    raw_max = np.max(iy)
    raw_argmax = np.argmax(iy)

    first_intercept = aw.attoworld_rs.find_first_intercept(iy, 0.5, 2)
    last_intercept = aw.attoworld_rs.find_last_intercept(iy, 0.5, 2)
    fwhm = aw.attoworld_rs.fwhm(iy, _dx)

    print(f"find_maximum_location gives maximum at ({interpolated_position},{interpolated_max})")
    print(f"raw indexing gives maximum at ({raw_argmax},{raw_max})")
    print(f"first intercept float index: {first_intercept}")
    print(f"fwhm (from function): {fwhm}")
    print(f"last intercept float index: {last_intercept}")

    aw.plot.showmo()
    return


@app.cell
def _():
    return


if __name__ == "__main__":
    app.run()
